/* iranalyse.cc
Analyse des Signals von IR-Fernsteuerungen
*/
#define VERSION "0.1" // 23.2.2014
/*

Prozessor: Atmega8 mit Quarz
Schaltung:
           LEDs an PB0-PB1
	   Taster an PD0-PD1
	   LCD-Modul PD2-PD7 (myAVR-Modul, 2-Zeilen)
	   TSOP an PC3

History:
23.2.2014     Erstellung

*/

#define F_CPU 3686400  //3.6864 MHz Quarz auf dem myAVR-MK2-Testboard
#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdio.h>
#define uchar unsigned char
#define uint unsigned int
#define ulong unsigned long

#define TASTE1 0
#define TASTE2 1

// Vordeklarationen:
void milliwait(int n);

/***************************** LCD-Modul ******************************/
#include "lcdmodul2.cc"

/*************************** Zeitbehandlung ***************************/
static volatile uint millisec=0, zaehler=0;
static volatile uchar achtelms=0,sec=0,min=0;

ISR(TIMER1_COMPA_vect)
{
#if(F_CPU==3686400)
 static uchar compzaehler=0;
 if(++compzaehler==5)
  {                   //Jedes 5. Mal durch 460 teilen, sonst durch 461,
   OCR1A=460;         //das ergibt dann durchschnittlich ein Teier von 460.8
   compzaehler=0;     //3686400/460.8 = 8000
  }
 else OCR1A=461;
#endif
 ++zaehler;
 if(++achtelms==8)
  {achtelms=0;
   if(++millisec==1000)
    {millisec=0;
     if(++sec==60)
      {sec=0;
       min++;
      }
    }
  }
}

void milliwait(int n) //n Millisekunden warten  1000 >= n >= 1
{
 uint ms;
 cli(); ms=millisec; sei();
 while(ms==millisec) ;//auf 1. Millisekunde warten
 if((ms+=n)>=1000) ms-=1000;
 while(ms!=millisec || ms!=millisec) //zweimal lesen um cli() sei() zu sparen
   ;//auf n. Millisekunde warten
}

void timer1_init()
{
 TCCR1B = (1<<WGM12)|(1<<CS10); //CTC-OCR1A-Modus, Prescaler=1
 OCR1A = F_CPU/8000; //falls dies nicht aufgeht, dann noch in der ISR korrigieren
 TCNT1 = 0; //Startwert des Timers
 TIMSK = 1<<OCIE1A; //Timer1 Interrupts bei Vergleichswert
}

/*************************** Hauptprogramm ****************************/
static uint zaehler1,zaehler2,bitlow[64],bithigh[64];
static char str[20];
static uchar nbits;

void code_berechnen(ulong* z1,ulong* z2)
{
 uchar bit,i,imax=nbits-1;
 *z1 = *z2 = 0;
 for(i=0;i<imax;i++)
  {
   if(bithigh[i] >= 2*bitlow[i]) bit=1; else bit=0;
   *z1 <<= 1;
   if((*z2&0x80000000)!=0) (*z1)++;
   *z2 = (*z2<<1)+bit;
  }
}

void pc_int0()
{
 uchar i;
 //Auf Prozessoren mit PCINT (Pin Change Interrupts) noch anzupassen
 //lcd_goto(1,1); lcd_write("IR-Signal erkannt"); //test
 cli(); zaehler=0; sei();
 while((PINC&(1<<3))==0) ;//warte bis Low-Puls fertig
 cli(); zaehler1=zaehler; zaehler=0; sei();
 while((PINC&(1<<3))!=0) ;//warte bis High-Puls fertig
 cli(); zaehler2=zaehler; zaehler=0; sei();
 for(i=0;i<64;)
  {
   while((PINC&(1<<3))==0) ;//warte bis Low-Puls fertig
   cli(); bitlow[i]=zaehler; zaehler=0; sei();
   while((PINC&(1<<3))!=0) ;//warte bis High-Puls fertig
   cli(); bithigh[i]=zaehler; zaehler=0; sei();
   i++;
   if(bithigh[i-1] > 4*zaehler1) break;
  }
 ulong z1,z2;
 nbits=i;
 milliwait(500);
 lcd_goto(1,1); lcd_write("Startbit  ");
 sprintf(str,"Low: %d ",zaehler1);
 lcd_goto(2,1); lcd_write(str);
 sprintf(str,"Hi: %d ",zaehler2);
 lcd_write(str);
 while((PIND&(1<<TASTE1))!=0) ;//warte auf Tastendruck
 lcd_goto(1,1); lcd_write("Anzahl Bits: ");
 sprintf(str,"%d ",nbits); lcd_write(str);
 code_berechnen(&z1,&z2);
 if(nbits-1>32) sprintf(str,"0x%lX%08lX  ",z1,z2);
 else           sprintf(str,"0x%08lX     ",z2);
 lcd_goto(2,1); lcd_write(str);
 milliwait(500);
 while((PIND&(1<<TASTE1)) && (PIND&(1<<TASTE2))!=0) ;//warte auf Tastendruck
 if((PIND&(1<<TASTE1))==0) //wenn Taste1 gedrueckt Werte fuer alle Bits anzeigen
  for(i=1;i<=nbits;)
   {
    sprintf(str,"%2d. Bit:        ",i);
    lcd_goto(1,1); lcd_write(str);
    sprintf(str,"Low:%3d Hi:%3d  ",bitlow[i-1],bithigh[i-1]);
    lcd_goto(2,1); lcd_write(str);
    milliwait(500);
    while((PIND&(1<<TASTE1)) && (PIND&(1<<TASTE2))!=0) ;//warte auf Tastendruck
    if((PIND&(1<<TASTE2))==0) {if(i>1) --i;} //mit Taste2 eins zurueck
    else i++; //sonst mit Taste1 eins weiter
    if(i==nbits) milliwait(500);
  }
 milliwait(999);
}

int main(void)
{
 DDRB = 0x03; //PB0, PB1 als Ausgang
 PORTD |= (1<<0); PORTD |= (1<<1); //Pullup-Widerstaende an PD0 und PD1
 PORTC |= (1<<3); //Pullup-Widerstand an PC3

 timer1_init();

 sei();
 
 lcd_init();
 lcd_write("IR-Analyse ");
 lcd_goto(2,1); //2.Zeile, 1.Spalte
 lcd_write("Version "); lcd_write(VERSION);
 
 while(1)
  {
   if((PINC&(1<<3))==0) pc_int0(); //Statt Verwendung von PCINT (da auf Atmega8 nicht vorhanden)
   PORTB = (PORTB&0xFC)+(sec&0x03); //test: Sekunden auf LEDs anzeigen
  }

 return 0;
}
