; lcdmodul.asm           letzte Aenderung: 16.9.2010
; Version: 0.0
; Prozessor: Atmega8
; Autor: Rolf Pfister
; Copyright: Freeware
; History:
; 27.8.2010	Erstellung
;
; Anwendung im Hauptprogramm:
; --------------------------
; .include "lcdmodul.asm"
; main:	rcall lcd_init
;
; oder bei Verwendung von Interrupts
; main:	rcall lcd_init1
;	sei
;	rcall lcd_init2
;-----------------------------------------------------------------------------
;; LCD-Modul auf oberen 6 Leitungen des entsprchenden PORTs
.equ LCD_PORT = PORTD
.equ LCD_DDR = DDRD
;-----------------------------------------------------------------------------
;.define EXTERNESMILLIWAIT	;falls milliwait1 schon vorhanden
.equ SRAMSTART=0x60	;beim Atmega8, ATmega32, ATtiny13A und ATtiny2313A
;.equ SRAMSTART=0x100	;beim ATmega328P
.equ PAGESIZEB = PAGESIZE*2 ; PAGESIZEB is page size in BYTES, not words
; Variablen im SRAM:
.equ lcdpos=SRAMSTART+PAGESIZEB		;LCD-Position
.equ lcdn0=SRAMSTART+PAGESIZEB+1
.equ lcdn1=SRAMSTART+PAGESIZEB+2
.equ lcdn2=SRAMSTART+PAGESIZEB+3
.equ lcdn3=SRAMSTART+PAGESIZEB+4	;Zahlen fuers LCD
.set SRAMFREI=SRAMSTART+PAGESIZEB+6  ;naechste freie gerade Adresse im SRAM
;-----------------------------------------------------------------------------
microwait_r16:			; etwa r16 usec warten
	push r16
	rjmp mic_L1
microwait1:			; etwa 1 usec warten
	push r16
	push r17
	ldi r16, 4
	rjmp mic_L2
microwait50:
	push r16
	ldi r16, 50
mic_L1: push r17
	add r16, r16
	add r16, r16
mic_L2:	dec r16
	brne mic_L2
	pop r17
	pop r16
	ret

milliwait5:			; etwa 5 Millisekunden warten
	ldi r16, 5
.ifdef EXTERNESMILLIWAIT
 	rjmp milliwait1
.else
milliwait1:			; r16 Anzahl Millisekunden warten
	push r16
	push r17
	push r18
mw_L0:	subi r16, 5
	brcs mw_L3
	ldi r17, high((F_CPU+500)/1000)
	ldi r18, low((F_CPU+500)/1000) ;Anzahl zu wartende Takte pro Millisec
mw_L1:	dec r18
	brne mw_L2
	dec r17
mw_L2:	brne mw_L1	; diese Schlaufe braucht 5 Takte, also 5 Millisec
	rjmp mw_L0
mw_L3:	subi r16, -5	;zuviel subtrahiertes wieder addieren
	breq mw_ret
mw_L4:	ldi r17, high((F_CPU+2500)/5000)
	ldi r18, low((F_CPU+2500)/5000) ;(Takte pro Millisec)/5
mw_L5:	dec r18
	brne mw_L6
	dec r17
mw_L6:	brne mw_L5	; diese Schlaufe braucht 5 Takte, also 1 Millisec
	dec r16
	brne mw_L4
mw_ret:	pop r18
	pop r17
	pop r16
	ret
.endif

lcd_cmd:	;void lcd_cmd(char cmd)
	cbi LCD_PORT, 2
	rcall microwait1
lcd_send:	; lcd_send(char data) ;r16=data
	push r17
	push r18
	in r17, LCD_PORT	;rs=PORTD
	andi r17, 4	;rs &= 4;
	ori r17, 3		; fuer Pullup-Widerstaende
	mov r18, r16	;char tmp=data
	andi r18, 0xF0	;tmp &= 0xF0
	or r18, r17	;tmp |= rs
	out LCD_PORT, r18	;PORTD=tmp
	rcall microwait1
	sbi LCD_PORT, 3	;PORTD |= (1<<3); //sbi(PORTD,3)
	rcall microwait1
	cbi LCD_PORT, 3	;PORTD &= ~(1<<3); //cbi(PORTD,3)
	mov r18, r16	;tmp=data
	andi r18, 0x0F	;tmp &= 0x0F
	swap r18	;tmp <<= 4
	or r18, r17	;tmp |= rs
	out LCD_PORT, r18	;PORTD=tmp
	rcall microwait1
	sbi LCD_PORT, 3	;PORTD |= (1<<3); //sbi(PORTD,3)
	rcall microwait1
	cbi LCD_PORT, 3	;PORTD &= ~(1<<3); //cbi(PORTD,3)
	ldi r16, 1
	rcall milliwait1
	pop r18
	pop r17
	ret

lcd_clear:
	push r16
	ldi r16, 0x01
	rcall lcd_cmd
	ldi r16, 2
	rcall milliwait1
	pop r16
	ret

lcd_home:
	push r16
	ldi r16, 0x02
	rcall lcd_cmd
	ldi r16, 2
	rcall milliwait1
	pop r16
	ret

lcd_on:	push r16
	ldi r16, 0x0E
	rcall lcd_cmd
	pop r16
	ret
lcd_off:
	push r16
	ldi r16, 0x08
	rcall lcd_cmd
	pop r16
	ret

lcd_init1:		;aufzurufen vor dem Einschalten der Interrupts
	push r16
	in r16, LCD_DDR
	ori r16, 0xFC	 ;nur obere 6 Bits veraendern
	out LCD_DDR, r16 ;DDRD=0xFC; //Port D obere 6 Bits auf Ausgang
	in r16, LCD_PORT
	andi r16, 0x03	 ;untere 2 Bits unveraendert lassen
	out LCD_PORT, r16
	pop r16
	ret
lcd_init:
	rcall lcd_init1
lcd_init2:		;aufzurufen nach dem Einschalten der Interrupts
	push r16
	push r17
	ldi r16, 50
	rcall milliwait1	;//Warten bis LCD-Controller gebootet hat
#define LCDWARMSTART
#ifdef LCDWARMSTART
	in r16, LCD_PORT
	andi r16, 0x03		;untere 2 Bits unveraendert lassen
	ori r16, 0x30		;Soft-Reset
	out LCD_PORT, r16	;PORTD=0x20; //4-Bit-Modus einschalten
	rcall microwait1
	sbi LCD_PORT, 3	;PORTD |= (1<<3); //sbi(PORTD,3); //Enable-Impuls
	rcall microwait1
	cbi LCD_PORT, 3	;PORTD &= ~(1<<3); //cbi(PORTD,3);
	rcall milliwait1
	in r16, LCD_PORT
	andi r16, 0x03		;untere 2 Bits unveraendert lassen
	ori r16, 0x30		;Soft-Reset 2.mal
	out LCD_PORT, r16	;PORTD=0x20; //4-Bit-Modus einschalten
	rcall microwait1
	sbi LCD_PORT, 3	;PORTD |= (1<<3); //sbi(PORTD,3); //Enable-Impuls
	rcall microwait1
	cbi LCD_PORT, 3	;PORTD &= ~(1<<3); //cbi(PORTD,3);
	rcall milliwait1
	in r16, LCD_PORT
	andi r16, 0x03		;untere 2 Bits unveraendert lassen
	ori r16, 0x30		;Soft-Reset 3.mal
	out LCD_PORT, r16	;PORTD=0x20; //4-Bit-Modus einschalten
	rcall microwait1
	sbi LCD_PORT, 3	;PORTD |= (1<<3); //sbi(PORTD,3); //Enable-Impuls
	rcall microwait1
	cbi LCD_PORT, 3	;PORTD &= ~(1<<3); //cbi(PORTD,3);
	rcall milliwait5
#endif
	in r16, LCD_PORT
	andi r16, 0x03		;untere 2 Bits unveraendert lassen
	ori r16, 0x20
	out LCD_PORT, r16	;PORTD=0x20; //4-Bit-Modus einschalten
	rcall microwait1
	sbi LCD_PORT, 3	;PORTD |= (1<<3); //sbi(PORTD,3); //Enable-Impuls
	rcall microwait1
	cbi LCD_PORT, 3	;PORTD &= ~(1<<3); //cbi(PORTD,3);
	rcall milliwait5
	ldi r16, 0x28	;lcd_cmd(0x28);
	rcall lcd_cmd	;  //Funktions-Set: 2 Zeilen, 5*7 Matrix, 4Bit
	rcall lcd_off
	rcall lcd_clear
	ldi r16, 0x06	;lcd_cmd(0x06); // Entry Mode
	rcall lcd_cmd
	rcall lcd_on
	ldi r16, 0x11
	sts lcdpos, r16	; Position 1.Zeile 1.Zeichen setzen
	pop r17
	pop r16
	ret

lcd_go:			;r16 obere 4 Bits=row, untere 4 Bits=col
	push r17
	mov r17, r16
	andi r17, 0x0F
	swap r16
	andi r16, 0x0F
	rcall lcd_goto
	pop r17
	ret
lcd_goto:	;r16=row=Zeile, r17=col=Spalte, r16,r17 zerstoert
	dec r16		;row--  Zaehlung statt bei 1 bei 0 beginnend
;	andi r16, 0x01	;row &= 0x01; //nur Zeile 0 oder Zeile 1 zulassen
	andi r16, 0x03	;auf 4 Zeilen beschraenkt
	swap r16
 	add r16, r16		
 	add r16, r16	;row <<= 6; //Zeile nach Bit 6 bringen
	dec r17		;col--;
;	andi r17, 0x0F	;col &= 0x0F; //nur Bereich 0 bis 15 zulassen
	andi r17, 0x3F	;auf 64 Zeichen Breite beschraenkt
	or r16, r17	;char tmp=row|col; //Adresse bilden
	ori r16, 0x80	;tmp |= 0x80; //Cursor setzen
	rcall lcd_cmd	;lcd_cmd(tmp);
	ret

lcd_write:	;void lcd_write(char r16=c)
	sbi LCD_PORT, 2
	rjmp lcd_send
lcd_print:	;Z=Zeiger im RAM auf einen String mit 0 terminiert
	ld r16, Z+
	tst r16
	breq lcd_p1
	rcall lcd_write
	rjmp lcd_print
lcd_p1:	ret

string_L1:
	rcall lcd_write
print_string:	    ;String in ZH:ZL auf lcd ausgeben, Z erhoeht, r16 zerst.
	lpm r16, Z+
	tst r16
	brne string_L1
	ret

lcd_r1:			; Zahl in r1 auf LCD anzeigen
	push r16
	push r17
;	ldi r16, 1
;	ldi r17, 1
;	rcall lcd_goto
	mov r16, r1
	andi r16, 0x0F
	ori r16, 0x30		; ASCII fuer Zahlen 0 bis 9
	rcall lcd_write
;	ldi r16, 2		;test
;	ldi r17, 1
;	rcall lcd_goto
;	mov r16, r1
;	inc r16			;test 1 mehr auf 2. Zeile
;	andi r16, 0x0F
;	ori r16, 0x30		; ASCII fuer Zahlen 0 bis 9
;	rcall lcd_write
	pop r17
	pop r16
	ret

lcd_r16r17:		; Zahl r16:r17 auf LCD anzeigen
	push r0
	push r2
	push r3
	push r20
	mov r3, r16
	mov r2, r17
	lds r16, lcdpos
;	cpi r16, 0x11		
;	brne lcd_L0
	rcall lcd_go
lcd_L0:	ldi r20, 10
	rcall divmodr20
	sts lcdn0, r0
	rcall divmodr20
	sts lcdn1, r0
	rcall divmodr20
	sts lcdn2, r0
	rcall divmodr20
	sts lcdn3, r0
	rcall divmodr20
	mov r16, r0
	tst r16
	breq lcd_L1
	ori r16, 0x30
	rjmp lcd_L2
lcd_L1:	ldi r16, 0x20		; ' '
lcd_L2:	rcall lcd_write
	lds r16, lcdn3
	ori r16, 0x30
	rcall lcd_write
	lds r16, lcdn2
	ori r16, 0x30
	rcall lcd_write
	lds r16, lcdn1
	ori r16, 0x30
	rcall lcd_write
	lds r16, lcdn0
	ori r16, 0x30
	rcall lcd_write
;	ldi r16, 0x20		; test
; 	rcall lcd_write		; test
	lds r16, lcdpos
	cpi r16, 0x11
	breq lcd_L11
	cpi r16, 0x18
	breq lcd_L18
	cpi r16, 0x21
	breq lcd_L21
	ldi r16, 0x11
	rjmp lcd_L3
lcd_L11:ldi r16, 0x18
	rjmp lcd_L3
lcd_L18:ldi r16, 0x21
	rjmp lcd_L3
lcd_L21:ldi r16, 0x28
lcd_L3:	sts lcdpos, r16
	pop r20
	pop r3
	pop r2
	pop r0
	ret
;-----------------------------------------------------------------------------
; divmodr20
;  16Bit vorzeichenlose Zahl dividieren durch 8Bit vorzeichenlosen Teiler
; Eingabe: r3:r2 = Zahl (16Bit), r20 = Teiler (8Bit) (unveraendert)
; Ausgabe: r3:r2 = Resultat, r0 = Rest
;-----------------------------------------------------------------------------
divmodr20:
	push r17
	clr r0
	ldi r17, 16
divL1:	lsl r2
	rol r3
	rol r0
	cp r0, r20
	brcs divL2
	sub r0, r20
	inc r2
divL2:	dec r17
	brne divL1
	pop r17
	ret

mul16:		;r3:r2:r1:r0 = r1:r0 * r3:r2
	push r16
	push r17
	push r18
	push r19
	push r4
	push r5
	mov r16, r0
	mov r17, r1		; AB
	mov r18, r2
	mov r19, r3		; CD
	mul r17, r19		;A*C
	mov r2, r0		; EFGH = AB*CD --> EF=A*C GH=B*D EFG+=A*D+B*C
	mov r3, r1		; r3r2 = EF=A*C
	mul r16, r18		;B*D
	mov r4, r0
	mov r5, r1		; r5r4 = GH=B*D
	mul r17, r18		;A*D
	clr r17
	add r5, r0		; EFG+=A*D
	adc r2, r1
	adc r3, r17
	mul r16, r19		;B*C
	add r5, r0		; EFG+=B*C
	adc r2, r1
	adc r3, r17
	mov r0, r4
	mov r1, r5
	pop r5
	pop r4
;pop19ret:
	pop r19
	pop r18
	pop r17
	pop r16
	ret

div16:			;r3:r2:r1:r0 /= r17:r16, r17:r16 = Rest
	push r18
	push r19
	push r20
	clr r18
	clr r19
	ldi r20, 32
div16L1:lsl r0
	rol r1
	rol r2
	rol r3
	rol r18
	rol r19
	cp  r18, r16
	cpc r19, r17
	brcs div16L2
	sub r18, r16
	sbc r19, r17
	inc r0
div16L2:dec r20
	brne div16L1
	mov r16, r18
	mov r17, r19		;r17:r16 = Rest
	pop r20
	pop r19
	pop r18
	ret
;------- Ende LCD-Anzeige ------------------------------------------------
