;-----------------------------------------------------------------------------
;* irempf13a.asm
;* Version: 0.02 	letzte Aenderung: 20.2.2014
;* Prozessor: ATtiny13A, interner Oszillator 1MHz
;* Schaltschema:
;*   irempf13a.png
;*
;* Autor: Rolf Pfister
;* Copyright: Freeware
;* History:
;* 29.7.2012	Erstellung
;* 4.8.12       Versuch PCINT3 zu verwenden
;* 23.3.2013	Fehler bei Benutzung von PCINT3 korrigiert: out GIMSK, r16 (nicht sts)
;* 18.2.2014	Versuch zu schalten bei Erkennung von Samsung-Startcode
;*
;*
;* globale Registerbelegung:
;* r10...r15 Zeiten (1msec bis Tage) fuer Uhr
;* r9 8tel Millisec fuer Uhr
;* r0 = SREG speichern im Timer-Interrupt
;* r8 = Erkennung von IR-Signal
;* r25:r24 Zeitdifferenz zwischen Flanken
;* r23:r22 empfangene IR-Bits
;* r21 = Samsung oder Kaon erkannt  (neu 18.2.2014, war r19 im irempfm8.asm, r19 aber auch in milliwait)
;*
;* global im Hauptprogramm:
;* r7 = Status
;*
;* Verwendung:
;*

;.equ FREQTEST=1
;.equ TIMERTEST=1
;.equ LEDTEST=1
;.equ RC5_CODE=1
;.equ SIRCS_CODE=1
.equ KAON_UND_SAMSUNG_CODE=1  	;neu 18.2.2014, von irempfm8.asm uebernommen
;.equ PCINT3_VERWENDEN=1

;-----------------------------------------------------------------------------
.include "tn13def.inc"		; Definitionsdatei fuer den Prozessortyp
; Reset und Interruptvectoren	;VNr. Beschreibung
begin:	rjmp	main		; 1   Power On Reset
	reti			; 2   IRQ0 Handler
.ifdef PCINT3_VERWENDEN
	rjmp	pc_int0		; 3   PCINT0
.else
	reti
.endif
	rjmp	tc0over		; 4   Timer0 Overflow
	reti			; 5   EEPROM ready
	reti			; 6   Analog Comperator
	reti			; 7   Timer0 CompareA
	reti			; 8   Timer0 CompareB
	reti			; 9   Watchdog Interrupt
	reti			; 10  ADC Conversion
;-----------------------------------------------------------------------------
; Variablen im SRAM:
.equ SRAMSTART=0x60
.equ richtung=SRAMSTART		;Motoransteuerung (0b01 oder 0b10)
.equ sensor1adwertL=SRAMSTART+1
.equ sensor1adwertH=SRAMSTART+2 ;Wert von AD-Wandler fuer Sensor

;-----------------------------------------------------------------------------
; Start, Power On, Reset
main:	ldi r16, low(RAMEND)
        out SPL, r16  	     ; Init Stackpointer

; Prescaler 8 sollte schon mit den Fuses gesetzt sein:
;	ldi r16, 0x80
;	out CLKPR, r16		; Vorbereitung zum Prescaler setzen
;	ldi r16, 3
;	out CLKPR, r16		; Clk-Prescaler auf 8 setzen --> 1MHz

	ldi r16, 0x65		; Wert ausprobieren fuer 1MHz
	out OSCCAL, r16		; Oszillator kallibrieren --> 1MHz

	ldi r16, 0x03		;PB0, PB1 auf Ausgang
        out DDRB, r16		;Data Direction Register B setzen
	ldi r16, 0x18		;Pullupwiderstaende PB4,PB3 fuer Taster und TSOP
	out PORTB, r16		;LEDs aus, Pullup-Widerstaende setzen

.ifdef FREQTEST
T1:	inc r16			; Test fuer Frequenzmessung
	out PORTB, r16		; Schlaufe braucht 4 Takte, eine Periode also 8
	rjmp T1			; Bei 1MHz muesste an B0 125kHz zu messen sein.
.endif

;; Timer Takt auf 1MHz und Teiler auf 125 setzen damit ein Interrupt alle 0.125msec (8kHz) erfolgt.
	ldi r17, 0x02		;Interruptenable setzen: B=8 A=4 OVR=2
	out TIMSK0, r17
	ldi r17, 0		;0 = Normal-Modus  (2 = CTC-Modus)
	out TCCR0A, r17		;Timer-Modus setzen
;; Vorteiler im TCCR0B: 1=nicht teilen, 2=durch 8, 3=64, 4=256, 5=1024
;	ldi r17, 2		;bei F_CPU==8MHz: Vorteiler auf 8 setzen (8MHz/8=1MHz)
	ldi r17, 1		;bei F_CPU==1MHz: nicht teilen
	out TCCR0B, r17
.equ	TCNTSTART=256-125
	ldi r17, TCNTSTART	;Startwert zum Teilen durch 125
	out TCNT0, r17		;Counter auf Startwert setzen

	ldi r16, 8
	mov r9, r16		;8tel Millisec auf Startwert setzen
	clr r10
	clr r11			; r11r10 = Millisec
	clr r12			; Sekunden
	clr r13			; Minuten
;	clr r14			; Stunden
;	clr r15			; Tage, auf 0 setzen

;; Interrupt fuer Aenderungen auf PB3 vorbereiten:
.ifdef PCINT3_VERWENDEN
	sbi PCMSK, PCINT3
	ldi r16, (1<<PCIE)
	out GIMSK, r16
.endif

	sei			;Interrupts erlauben

.ifdef TIMERTEST
	ldi r17, 20		;erste 20 Sekunden
T1:	mov r16, r12
	ori r16, 0x18		;Pullupwiderstaende setzen
	out PORTB, r12		; Sekunden auf LEDs
	cp r12, r17
	brne T1
	ldi r16, 0x18		;Pullupwiderstaende PB4,PB3 fuer Taster und TSOP
	out PORTB, r16		;LEDs aus, Pullup-Widerstaende setzen
.endif

;-----------------------------------------------------------------------------
	ldi r21, 0		; EmpfangsStatus: 0 = warte auf Startbit
	clr r7			; Ausgangs-Status loeschen
	clr r8			; Flag fuer IR-Signal loeschen
;	in r16, PINB		; Eingaenge lesen
;	lsl r16
;	swap r16		; PB4,PB3 auf Bits 1,0 bringen
;	andi r16, 0x03		; restliche Bits loeschen

;; normaler Hauptloop
mainloop:
.ifdef LEDEST
;; Test: LED jede Sekunde kurz aufleuchten lassen:
	mov r16, r12
_m3:	cp r16, r12
	breq _m3		; auf neue Sekunde warten
	sbi PORTB, 1		; rote LED ein
	rcall waitfuenftelsec
	cbi PORTB, 1		; rote LED aus
.else
	ldi r16, 1
	clr r17
	rcall milliwait
.endif
;; Sleep-Modus (kein Unterschied messbar)
;	cbi ADCSRA, ADEN        ; AD-Wandler ausschalten
;	ldi r16, (1<<SE)	;Idle
;	out MCUCR, r16
;	sleep

;; Ausgang ein/aus-schalten mit Taster:
	in r16, PINB		; Eingaenge lesen
	andi r16, 0x10		; PB4 testen
	brne ircheck		; gesetzt --> Taste nicht gedrueckt
status_wechseln:
	inc r7			; Ausgangs-Status wechseln
	mov r16, r7
	andi r16, 1
	breq ausgang_aus
ausgang_ein:
	sbi PORTB, 0		; Ausgang ein
	rjmp blink
ausgang_aus:
	cbi PORTB, 0		; Ausgang aus
blink:	sbi PORTB, 1		; rote LED ein
	rcall waitfuenftelsec
	cbi PORTB, 1		; rote LED aus
	rcall waitfuenftelsec
	rjmp mainloop

;; Ausgang ein/aus-schalten mit IR-Fernsteuerung:
ircheck:
.ifdef PCINT3_VERWENDEN
.else
	sbis PINB, 3		; PB3 auf high?
	rcall pc_int0		; nein: Interruptroutine aufrufen
.endif
	mov r16, r8		; IR-Signal von Interruptroutine
	tst r16			; erkannt?
	breq irno		; nein -->
	clr r8
.ifdef PCINT3_VERWENDEN
	sbi PCMSK, PCINT3	; Interrupt fuer naechstes Startbit wieder einschalten
.endif
	rjmp status_wechseln	; bei Erfolg --> ein/aus-schalten
irno:	rjmp mainloop		; sonst weiter mit Hauptschlaufe

;stopp:	cli			; Interrupts aus
;	ldi r16, 0		; alle LEDs aus
;	out PORTB, r16
;	out DDRB, r16		; alle Ports auf Eingang
;	ldi r16, (1<<SE)|(1<<SM1) ;Power-down setzen
;	out MCUCR, r16
;	sleep
;	rjmp main		;sollte nie erreicht werden.

waitfuenftelsec:
	push r17
	ldi r16, low(200)
	ldi r17, high(200)
	rcall milliwait		; 0.2 Sec
	pop r17
	ret

;wait2einhalb:			; 2.5 sec warten, r16r17 zerstoert
;	ldi r16, low(500)
;	ldi r17, high(500)
;	rcall milliwait		; 0.5 Sec
;	ldi r16, 2		; 2 Sec
;	rjmp secwait

;secwait:	; r16 Sekunden lang warten
;	push r17
;	push r18
;	mov r18, r16
;	add r18, r16
;_sw1:	ldi r16, low(500)
;	ldi r17, high(500)
;	rcall milliwait
;	dec r18
;	brne _sw1
;	pop r18
;	pop r17
;	ret

;-----------------------------------------------------------------------------
; OVR-Interrupt-Routine, wird alle 0.125 Millisekunden aufgerufen
; Registerbelegungen:
; r9 8tel Millisekunden, r11:r10 Millisekunden, r12 Sekunden, r13 Minuten, r14 Stunden, r15 Tage
tc0over:push r16
	ldi r16, TCNTSTART	;fuer Normal-Modus
	out TCNT0, r16		;Counter auf Startwert setzen
	in r0, SREG	;Statusregister sichern
	adiw r25:r24, 1	;Zeitdifferenzen in 1/8-Microsekunden
	brne L0
	ldi r25, 0xFF	;bei Ueberlauf r25:r24 auf hohen Wert setzen
L0:	dec r9
	brne end2
	ldi r16, 8
	mov r9, r16
millisekzaehlen:
	push r17
	inc r10		;Millisekunden erhoehen
	brne L1
	inc r11
L1:	ldi r16, low(1000)
	ldi r17, high(1000)	;r17:r16 = 1000
	cp r10, r16
	cpc r11, r17
	brne end1	;fertig wenn msec!=1000
	clr r10
	clr r11		;Millisekunden auf 0 setzen
	inc r12		;Sekunden erhoehen
	ldi r16, 60
	cp r12, r16
	brne end1	;fertig wenn sec!=60
	clr r12		;Sekunden auf 0 setzen
	inc r13		;Minuten erhoehen
;	cp r13, r16
;	brne end1	;fertig wenn min!=60
;	clr r13		;Minuten auf 0 setzen
;	inc r14		;Stunden erhoehen
;	ldi r16, 24
;	cp r14, r16
;	brne end1	;fertig wenn std!=24
;	clr r14		;Stunden auf 0 setzen
;	inc r15		;Tage erhoehen
end1:	pop r17
end2:	out SREG, r0	;Statusregister zurueckholen
	pop r16
	reti

;-----------------------------------------------------------------------------
; PCINT0-Interrupt-Routine, bei Aenderungen von PB3 (PCINT3) aufgerufen
; (bei Version mit Atmega8 bei Aenderung von PC3 vom Hauptprogramm aus aufgerufen)

.ifdef KAON_UND_SAMSUNG_CODE

.equ KAON_STARTL=64		;Startbit 8ms auf Low (Werte in 1/8 Millisek)
.equ KAON_STARTH=32		;Startbit 4ms auf High
.equ SAMSUNG_STARTL=32		;Startbit 4ms auf Low
.equ SAMSUNG_STARTH=32		;Startbit 4ms auf High
pc_int0:
	push r16
	in r16, SREG	;Statusregister sichern
.ifdef PCINT3_VERWENDEN
	cbi PCMSK, PCINT3	; Interrupt fuer steigende Flanke abschalten
	sei
.endif
	push r16
	push r17
;	rjmp pc_code_erkannt	;test
	sbic PINB, 3		; PB3 auf low?
	rjmp pc_fehler		; nein->
	tst r21			;ist es das Startbit?
	brne pc_L1		;nein-->
	clr r25			;ja: r25:24 ruecksetzen und
	clr r24			;dann Laenge des Startbits messen
pc_w1:	sbis PINB, 3		; PB3 auf high?
	rjmp pc_w1		; nein-> warten bis high
	tst r25
	brne startbit_falsch	  ;Starbit zu lang
	cpi r24, SAMSUNG_STARTL-4 ;kleiner als 0.5 ms weniger als erwartete Startbitlaenge?
	brcs startbit_falsch	  ;ja-> Startbit zu kurz
	cpi r24, SAMSUNG_STARTL+4 ;Startbitlaenge +- 0.5 ms von erwarteter fuer SAMSUNG-Code?
	brcs samsung_startbit	  ;ja-> Startbit erkannt
	cpi r24, KAON_STARTL-4 	  ;kleiner als 0.5 ms weniger als erwartet fuer KAON?
	brcs startbit_falsch	  ;ja-> Startbit zu kurz
	cpi r24, KAON_STARTL+4 	  ;Startbitlaenge +- 0.5 ms von erwarteter fuer KAON-Code?
	brcs kaon_startbit	  ;ja-> Startbit erkannt
startbit_falsch:
	ldi r16, low(500)
	ldi r17, high(500)
	rcall milliwait		;Warten bis aktuelle Pulse vorbei
	clr r21
	rjmp pc_int0_fertig
samsung_startbit:
	ldi r21, 1		;1 = SAMSUNG
;	rjmp pc_int0_fertig
	rjmp pc_L1		;provisorisch
kaon_startbit:
	ldi r21, 2		;2 = KAON
	rjmp pc_int0_fertig
pc_L1:	nop			;hier noch programmieren zum Code weiter auswerten
	clr r21			;Status wieder auf 0 fuer naechste Erkennung
pc_code_erkannt:
	inc r8			; Flag fuer erfolgreichen IR-Empfang setzen
pc_int0_fertig:
	clr r25
	clr r24			;Abstandzaehler ruecksetzen
pc_int0_ret:
	pop r17
	pop r16
.ifdef PCINT3_VERWENDEN
	cli
	tst r8
	brne pc_sre
	sbi PCMSK, PCINT3	; Interrupt fuer fallende Flanke wieder einschalten
.endif
pc_sre:	out SREG, r16	;Statusregister zurueckholen
	pop r16
	reti
pc_fehler:			;sollte nicht passieren
	ldi r17, 5		; 5 mal LED blinken lassen
pc_f1:	sbi PORTB, 1		; rote LED ein
	rcall waitfuenftelsec
	cbi PORTB, 1		; rote LED aus
	rcall waitfuenftelsec
	dec r17
	brne pc_f1
	rjmp pc_int0_ret
	
.else
; PCINT0-Interrupt-Routine, Version fuer RC5 und SIRCS
	
.ifdef RC5_CODE
.equ RC5_MAX=28  		;Maximale Laenge innerhalb eines Packets
.equ RC5_LANG=10		;Minimale Laenge fuer Langer Abstand
.equ RC5_SCHALTCODE=0x1234	;provi.
.endif
.ifdef SIRCS_CODE
.equ SIRCS_START_MIN=17		;Minimale Laenge des Startbits
.equ SIRCS_START_MAX=21		;Maximale Laenge des Startbits
.equ SIRCS_LUECKE=5		;ungefaehre Laenge der Luecke zwischen den Bits
.equ SIRCS_BIT0=5		;ungefaehre Laenge eines 0-Bits
.equ SIRCS_BIT1=10		;ungefaehre Laenge eines 1-Bits
.equ SIRCS_NBITS=32		;Anzahl Bits ohne Startbit
.endif

pc_int0:
	push r16
	in r16, SREG	;Statusregister sichern
	push r16
	push r17
;	rjmp pc_code_erkannt	;test
.ifdef RC5_CODE
	ldi r16, low(RC5_MAX)
	ldi r17, high(RC5_MAX)
	cp r24, r16
	cpc r25, r17
	brcc rc5_packet_start
	mov r16, r5
	cpi r16, 27
	brcc pc_int0_fertig	;allerletztes Halbbit ignorieren
	breq rc5_packet_start
	ldi r16, low(RC5_LANG)
	ldi r17, high(RC5_LANG)
	cp r24, r16
	cpc r25, r17
	brcc rc5lang
	rjmp rc5kurz
rc5_packet_start:
	sbic PINB, 3		; PB3 auf low?
	rjmp pc_int0_ret	; nein:	fehlerhaftes Startbit ignorieren
	clr r5
	inc r5			;Zaehler fuer Halbbits auf 1 setzen
	clr r22
	clr r23			;empfangene Bits loeschen
	rjmp pc_int0_fertig
rc5lang:
	inc r5			;naechstes Halbbit
	sbrc r5, 0		;Halbbit ungerade? (Skip wenn in r5 bit 0 geloescht ist)
	rcall bit_speichern	;ja: Bit speichern
rc5kurz:
	inc r5			;naechstes Halbbit
	sbrc r5, 0		;Halbbit ungerade? (Skip wenn in r5 bit 0 geloescht ist)
	rcall bit_speichern	;ja: Bit speichern
	mov r16, r5
	cpi r16, 27		;13 Bits (ohne Startbit) erreicht?
	brcs pc_int0_fertig	;nein-->
;; hier muesste noch RC5-Code ausgewertet werden
;	ldi r16, low(RC5_SCHALTCODE)
;	ldi r17, high(RC5_SCHALTCODE)
	mov r16, r22		;provi.
	mov r17, r23		;provi.
	cp  r22, r16
	cpc r23, r17
	breq pc_code_erkannt
	rjmp pc_int0_fertig
.endif
.ifdef SIRCS_CODE
	sbic PINB, 3		; PB3 auf low?
	rjmp pc_int0_fertig	; ja--> Start eines Bits
	ldi r16, low(SIRCS_START_MIN)
	ldi r17, high(SIRCS_START_MIN)
	cp r24, r16
	cpc r25, r17
	brcc sircs_startbit
;; hier die Bits auswerten:
	mov r16, r5
	cpi r16, SIRCS_NBITS	;Anzahl Bits (ohne Startbit) erreicht?
	brcc pc_int0_ret	;ja--> zu viele Bits ignorieren
	ldi r16, SIRCS_BIT0+2
	ldi r17, 0
	cp r24, r16
	cpc r25, r17		; langes Bit?
	brcs sircs_L1		; nein--> neues Bit = 0
	ldi r17, 1		; ja: neues Bit = 1
sircs_L1:
	ror r17			; neues Bit ins Cy schieben
	rol r22
	rol r23			; Neues Bit in r23:r22 schieben
	inc r5			; Bitzaehler erhoehen
	mov r16, r5
	cpi r16, SIRCS_NBITS	;Anzahl Bits (ohne Startbit) erreicht?
	brcs pc_int0_fertig	;nein-->

;; hier muessten noch die SIRCS-Codes ausgewertet werden
	rjmp pc_int0_fertig
	
sircs_startbit:
	ldi r16, low(SIRCS_START_MAX)
	ldi r17, high(SIRCS_START_MAX)
	cp r24, r16
	cpc r25, r17
	brcc pc_int0_ret	; zu langes Startbit ignorieren
	clr r5			; Bitzahler loeschen
	rjmp pc_code_erkannt	; provisorisch:	Startbit als erfolgreicher IR-Empfang ansehen
.endif

pc_code_erkannt:
	inc r8			; Flag fuer erfolgreichen IR-Empfang setzen
pc_int0_fertig:
	clr r25
	clr r24			;Abstandzaehler ruecksetzen
pc_int0_ret:
	pop r17
	pop r16
	out SREG, r16	;Statusregister zurueckholen
	pop r16
	reti

.endif
;-----------------------------------------------------------------------------
;; Routinen fuer Aufruf aus pc_int0
bit_speichern:
	clc			; Cy loeschen
	sbis PINB, 3		; PB3 gesetzt?
	sec			; nein: Cy setzen
	rol r22
	rol r23			; Neues Bit in r23:r22 schieben
	ret
	
;-----------------------------------------------------------------------------
milliwait:			;Wartezeit: r17:r16 msec, min 1 max 999
	push r18
	push r19
	push r16
	push r17
	cli
	mov r18, r10
	mov r19, r11		;r19r18 = sollzeit = aktuelle Millisec
	sei
	add r18, r16
	adc r19, r17		;sollzeit += Wartezeit
	ldi r16, 0xE8		;auf 1000 testen
	cpi r19, 0x03		;oberes Byte testen
	brcs loop1		;sollzeit<1000 --> weiter
	brne subtra		;sollzeit>1000 --> subtrahiere 1000
	cp r18, r16		;unteres Byte testen
	brcs loop1		;sollzeit<1000 --> weiter
subtra:	sub r18, r16
	sbci r19, 0x03		;wenn sollzeit>=1000 dann sollzeit-=1000
loop1:	cp r18, r10
	brne loop1
	cp r19, r11		;auf sollzeit warten
	brne loop1
	pop r17
	pop r16
pop1918ret:
	pop r19
	pop r18
	ret

;-----------------------------------------------------------------------------
.ifdef MICROWAIT_VERWENDEN
microwait:      ;; genau r16 usec warten, bei 1MHz Takt
		;; mindestens 22, maximal 255
	subi r16, 18		; Anzahl Takte ausserhalb der Schlaufe subtrahieren
	push r16
	andi r16, 3		; Rest beim Dividieren durch 4
	breq mic_r0		; Rest==0 --> kein zusaetzlicher Takt
	dec r16
	breq mic_r1		; Rest==1 --> 1 zusaetzlicher Takt
	dec r16
	breq mic_r2		; Rest==2 --> 2 zusaetzliche Takte
	rjmp mic_r3		; sonst Rest==3 --> 3 zusaetzliche Takte
mic_r0:	nop
mic_r1:	nop
mic_r2:
mic_r3:	pop r16
	lsr r16
	lsr r16			; r16 /= 4
mic_L4:	nop
	dec r16
	brne mic_L4		; Schlaufe braucht 4usec
	ret
.endif

;-----------------------------------------------------------------------------
.ifdef TEMPSENSOR
tabelle:	;static int tabelle[]=
	;; Temparaturen 10*mehr damit wir keine Fliesszahlen brauchen
	;; -550 ist also -55.0 GradCelsius
	.dw -550, 990, -500, 1040, -400, 1146, -300, 1260, -200, 1381
	.dw -100, 1510, 0, 1646, 100, 1790, 200, 1941, 250, 2020
	.dw 300, 2100, 400, 2267, 500, 2441, 600, 2623, 700, 2812
	.dw 800, 3009, 900, 3214, 1000, 3426, 1100, 3643, 1200, 3855
	.dw 1250, 3955, 1300, 4048, 1400, 4208, 1500, 4323, 0, 0
.endif
