/* test1.cc       letzte Aenderung: 30.6.2012

 Prozessor: ATmega8, ATmega32
 Schaltung: Port B0 mit Frequenzmesser oder Oszilloskop verbunden
                 B1 mit LED
                 B2 mit IR-LED
 Autor: Rolf Pfister
 Copyright: Freeware
 History:
 29.6.2012	Erstellung
*/

#define TIMER1_VARIANTE //geht gerade so mit 8MHz Quarz fuer bis 38kHz Takt
//#define TIMER0_VARIANTE //sehr ungenau da Takte in der ISR relevant sind

#include <avr/io.h>
#include <avr/interrupt.h>
#include "ulong.h"

#define TAKT_BIT 0
#define LED1_BIT 1
#define IRLED_BIT 2
#define LED2_BIT 3

#define F_TAKT 38  //gewuenschte Frequenz in kHz
//#define F_TAKT 19  //gewuenschte Frequenz in kHz

/**** Takterzeugung ****/
static volatile uchar taktzaehler=0,sendeflag=0;

// Genaue Quarzfrequenz in Hertz:
#define F_CPU 16000000L
//#define F_CPU 3686400L

#define NormalerCompwert (((F_CPU/F_TAKT+500)/2000)-1) //bei weniger 0.75 abgerundeter Wert
#define ZweiterCompwert (((F_CPU/F_TAKT+1500)/2000)-1) //bei mehr als 0.25 aufgerundeter Wert
/* Beispiel:
80000000 / 38 / 2000 = 105.263
abgerundet:  105
aufgerundet: 106
Durchschnitt: 105.5
Abweichung 105.5/105.263 --> 0.225%
*/

#ifdef TIMER1_VARIANTE
ISR(SIG_OUTPUT_COMPARE1A)
{
 static uchar takt=1;
 taktzaehler++;

 if((takt^=1)!=0)
  {
   OCR1A=NormalerCompwert;
   PORTB |= (1<<TAKT_BIT); //PB0 setzen
   if(sendeflag!=0) PORTB |= (1<<IRLED_BIT); //falls sendeflag aktiv: IR-LED einschalten
  }
 else
  {
   OCR1A=ZweiterCompwert;
   PORTB &= ~(1<<TAKT_BIT); //PB0 loeschen
   PORTB &= ~(1<<IRLED_BIT); //IR-LED ausschalten
  }
}

void timer1_init()
{
 TCCR1B = (1<<WGM12)|(1<<CS10); //CTC-OCR1A-Modus, Prescaler=1 (also Timer-Takt ist gleich Quarz-Takt)
 OCR1A = NormalerCompwert;
 TCNT1 = 0; //Startwert des Timers
 TIMSK = 1<<OCIE1A; //Timer1 Interrupts bei Vergleichswert
}

#else
//#ifdef TIMER0_VARIANTE

ISR(TIMER0_OVF_vect)
{
 static uchar takt=1;
 TCNT0=256-NormalerCompwert;
 //taktzaehler++;

 if((takt^=1)!=0)
  {
   PORTB |= (1<<TAKT_BIT); //PB0 setzen
   //if(sendeflag!=0) PORTB |= (1<<IRLED_BIT); //falls sendeflag aktiv: IR-LED einschalten
  }
 else
  {
   PORTB &= ~(1<<TAKT_BIT); //PB0 loeschen
   //PORTB &= ~(1<<IRLED_BIT); //IR-LED ausschalten
  }
}

void timer0_init()
{
 TCCR0 = (1<<CS00); //CTC-OCR1A-Modus, Prescaler=1 (also Timer-Takt ist gleich Quarz-Takt)
 TCNT0 = 256-NormalerCompwert;//Startwert des Timers
 TIMSK = (1<<TOIE0); //Timer0 Interrupts bei Ueberlauf
}

#endif

//Bei einem unerwarteten Interrupt nichts machen:
void __attribute__ ((naked)) __vector_default(void) {__asm__ ("reti");}

int main(void)
{
 uchar n=0;
 DDRB  = 0x07; //PortB PB0-2 als Ausgaenge
 PORTB = n; //alle Auasgaenge aus

#ifdef TIMER0_VARIANTE
 timer0_init();
#else
 timer1_init();
#endif

 sei(); //Interrupt einschalten

 while(1) //Hauptschlaufe
  {
   if(taktzaehler>=42) //Blinkfrequenz 42 mal kleiner als Takt
    {
     cli();
     taktzaehler -= 42;
     sei();
     //LED und IR-LED blinken lassen:
     n ^= 6;
     if(n&2) PORTB |= (1<<LED1_BIT);
     else    PORTB &= ~(1<<LED1_BIT);
     sendeflag = n&4;
     if(n&4) PORTB |= (1<<IRLED_BIT);
     else    PORTB &= ~(1<<IRLED_BIT);
    }
  }

 return 0;
}
